require File.expand_path( File.join(File.dirname(__FILE__), "..", "beanpicker") )
require File.expand_path( File.join(File.dirname(__FILE__), "process") )

# Used by Beanpicker to kill childs and bury jobs
BEANPICKER_FORK = { :child_every => false, :child_master => false, :child_pid => 0 }

module Beanpicker

  # Used by combine to run a server of jobs
  class Server

    include MsgLogger

    # expect a list of files to load or pick from ARGV
    def initialize(args=ARGV)
      @args = args
    end

    # Create all workers and verify if exists any Worker or Worker::Child to start a loop.
    # If no only exit
    def run
      debug "Hiring workers..."
      for arg in @args
        Worker.new arg
      end

      w = c = 0
      for worker in Beanpicker::workers
        w += 1
        m = "Worker: #{worker.name} hired to do"
        f = {}
        for child in worker.childs
          c += 1
          f[child.job_name] ||= 0
          f[child.job_name] += 1
        end
        for job in f.keys.sort
          m << " #{job}##{f[job]}"
        end
        debug m
      end

      if w == 0
        fatal ["NOBODY WANT TO WORK TO YOU!!",
               "Have you specified a file?",
               "e.g. #{$0} lib/my_file_with_jobs.rb"].join("\n")
        exit
      end

      if c == 0
        fatal ["ALL YOUR #{w} WORKER#{"S" if w > 1} ARE LAZY AND DON'T WANT TO DO ANYTHING!!",
               "Have you specified a job in any of yous files?",
               "e.g. job('no.lazy.worker') { |args| puts args[:my_money] }"].join("\n")
        exit
      end

      debug "#{w} worker#{"s" if w > 1} hired to do #{c} job#{"s" if c > 1}, counting the money..."
      
      loop { sleep 1 }
    end

    # Use the Beanpicker::log_handler to log messages
    def log_handler
      Beanpicker::log_handler
    end

  end

  # Worker is the class used to create jobs
  class Worker

    include MsgLogger

    # The name of worker. Pick from args[:name] or from the name of file(if passed)
    attr_reader :name
    # The list of childs
    attr_reader :childs

    # Accept a file and/or a block with jobs and add itself to workers list
    def initialize(filepath=nil, args={}, &blk)
      @childs = []
      @name   = args[:name] rescue nil
      if filepath
        @name = filepath.split(/[\\\/]/)[-1].gsub(/\.[^\.]+$/,'').split(/[_\.]/).map do |x|
          x.capitalize
        end.join if @name.nil?

        begin
          instance_eval File.read(filepath)
        rescue => e
          error Beanpicker::exception_message(e, "when loading file #{filepath}")
        end
      end

      if block_given?
        begin
          instance_eval(&blk)
        rescue => e
          error Beanpicker::exception_message(e, "when evaluating block")
        end
      end

      @name = "BeanpickerWorker without name" unless @name
      Beanpicker::add_worker(self)
    end

    # Pass the job to Child::process and add the return to childs list
    def job(name, args={}, &blk)
      @childs << Child.process(name, args, self, &blk)
      @childs.flatten!
    end

    # Use it own log_handler or global if not defined
    def log_handler
      @log_handler || Beanpicker::log_handler
    end

    # Shortcut to log_handler=
    def log_file(f)
      #without self call don't call log_handler= of this class Oo
      self.log_handler = f
    end

    # Child is the class that handle the job.
    # Every job can be one or more childs
    class Child

      include MsgLogger

      # Process job.
      # Use opts[:childs] or Beanpicker::default_childs_number to determine how many childs should be created
      def self.process(job, opts={}, worker=nil, &blk)
        (opts[:childs] || Beanpicker::default_childs_number).times.map do |i|
          Child.new(job, opts, i, worker, &blk)
        end
      end


      # The name of job
      attr_reader :job_name
      # The number of job(generated by Child::process)
      attr_reader :number
      # Should fork every job?
      attr_reader :fork_every
      # Should fork the child process?
      attr_reader :fork_master
      # The pid of running job
      attr_reader :fork_every_pid
      # The pid of forked child process
      attr_reader :fork_master_pid
      # The default options merged with argument options
      attr_reader :opts
      # The Worker (father?) or nil
      attr_reader :worker

      # Create a new job, start a fork if @opts[:fork_master] and start the work
      def initialize(job, opts={}, number=0, worker=nil, &blk)
        @job_name    = job
        @opts        = {
          :childs      => Beanpicker::default_childs_number,
          :fork_every  => Beanpicker::default_fork_every,
          :fork_master => Beanpicker::default_fork_master
        }.merge(opts)
        @number      = number
        @blk         = blk
        @loop        = nil
        @beanstalk   = Beanpicker::new_beanstalk
        @run         = true
        @job         = nil
        @worker      = worker
        if @opts[:fork]
          _fork_every  = @opts[:fork].to_s == 'every'
          _fork_master = @opts[:fork].to_s == 'master'
        else
          _fork_every  = !!@opts[:fork_every]
          _fork_master = !!@opts[:fork_master]
        end
        #really need self
        self.log_handler = @opts[:log_file] unless @opts[:log_file].nil?
        @fork_every  = Beanpicker::fork_every.nil?  ? _fork_every  : Beanpicker::fork_every
        @fork_master = Beanpicker::fork_master.nil? ? _fork_master : Beanpicker::fork_master
        @fork_master_pid = nil
        @fork_every_pid  = nil
        start_watch
        start_loop
      end

      # Return the beanstalk connection(Child don't use the global connection)
      def beanstalk
        @beanstalk
      end

      # Watch the tube with job name and ignore all others
      def start_watch
        beanstalk.watch(@job_name)
        beanstalk.list_tubes_watched.each do |server, tubes|
          tubes.each { |tube| beanstalk.ignore(tube) unless tube == @job_name }
        end
      end

      # Start the loop, fork if needed
      def start_loop
        return false if @loop and @loop.alive?
        if @fork_master
          fork_master_child_and_monitor
        else
          @loop = Thread.new(self) do |child|
            work_loop(child)
          end
        end
      end

      # call start_work passing itself(child) while @run
      def work_loop(child)
        start_work(child) while @run
      end

      # Here is all the magic :)
      #
      # Call fork and start the job and delete on end, bury the job if a error is raised
      def start_work(child=self)
        fork do
          begin
            @job = child.beanstalk.reserve
            BEANPICKER_FORK[:job] = @job if BEANPICKER_FORK[:child_every]
            data  = @job.ybody

            if not data.is_a?(Hash) or [:args, :next_jobs] - data.keys != []
              data = { :args => data, :next_jobs => [] }
            end

            t=Time.now
            debug "Running #{@job_name}##{@number} with args #{data[:args]}; next jobs #{data[:next_jobs]}"
            r = @blk.call(data[:args].clone)
            debug "Job #{@job_name}##{@number} finished in #{Time.now-t} seconds with return #{r}"
            data[:args].merge!(r) if r.is_a?(Hash) and data[:args].is_a?(Hash)

            @job.delete

            Beanpicker.enqueue(data[:next_jobs], data[:args]) if r and not data[:next_jobs].empty?
          rescue => e
            fatal Beanpicker::exception_message(e, "in loop of #{@job_name}##{@number} with pid #{Process.pid}")
            if BEANPICKER_FORK[:child_every]
              exit
            else
              Thread.new(@job) { |j| j.bury rescue nil }
            end
          ensure
            @job = nil
          end
        end
      end

      # Fork the process if @fork_every and wait or only call the block
      def fork(&blk)
        if @fork_every
          @fork_every_pid = pid = Kernel.fork do
            BEANPICKER_FORK[:child_every] = true
            Process.die_with_parent
            at_exit_to_every_child_fork
            $0 = "Beanpicker job child #{@job_name}##{@number} of #{Process.ppid}"
            blk.call
          end
          if BEANPICKER_FORK[:child_master]
            BEANPICKER_FORK[:child_pid] = pid
            Process.waitpid pid
            BEANPICKER_FORK[:child_pid] = nil
          else
            Process.waitpid pid
          end
          @fork_every_pid = nil
        else
          blk.call
        end
      end

      # The child still working?
      def running?
        @run
      end

      # Stop running, kill the thread and kill master os every process
      def die!
        @run = false
        @loop.kill if @loop and @loop.alive?

        kill_pid = nil
        if @fork_master
          kill_pid = @fork_master_pid
        elsif @fork_every
          kill_pid = @fork_every_pid
        end

        if kill_pid and kill_pid.is_a?(Integer) and Process.running?(kill_pid)
          debug "Killing child with pid #{kill_pid}"
          Process.kill "TERM", kill_pid
        end

      end

      # Crete a new fork, change the name and create a thread to restart the fork if it die
      #
      # Called by start_loop when fork_master
      def fork_master_child_and_monitor
        @fork_master_pid = Kernel.fork do
          at_exit_to_master_child_fork
          Process.die_with_parent
          BEANPICKER_FORK[:child_master] = true
          $0 = "Beanpicker master child #{@job_name}##{@number}"
          work_loop(self)
        end
        @loop = Thread.new(self) do |child|
          Process.waitpid @fork_master_pid
          child.fork_master_child_and_monitor if child.running?
        end
      end

      # Create a at_exit to kill the child(if any)
      #
      # Used by fork_master_child_and_monitor
      def at_exit_to_master_child_fork
        at_exit do
          pid = BEANPICKER_FORK[:child_pid]
          if pid and pid > 0
            if Process.running?(pid)
              Process.kill "TERM", pid
              sleep 0.1
              if Process.running?(pid)
                sleep 2
                Process.kill "KILL", pid if Process.running?(pid)
              end
            end
          end
          Kernel.exit!
        end
      end

      # Create a at_exit to bury the job(if any) and exit
      #
      # Used by fork
      def at_exit_to_every_child_fork
        at_exit do
          Thread.new do
            sleep 1
            Kernel.exit!
          end
          BEANPICKER_FORK[:job].bury rescue nil if BEANPICKER_FORK[:job]
        end
      end

      # Return the own log_handler, or the Worker log_handler or the global log_handler
      def log_handler
        #'@log_handler || ' go to worker/global log_handler even if @log_handler is defined
        defined?(@log_handler) ? @log_handler : @worker.nil? ? Beanpicker::log_handler : @worker.log_handler
      end

    end

  end

end

# kill childs or stop jobs
at_exit do
  if not BEANPICKER_FORK[:child_master] and not BEANPICKER_FORK[:child_every]
    Beanpicker::debug "Laying off workers..." if Beanpicker::workers.count > 0
    Beanpicker::stop_workers
  end
end

# hide errors throwed with ctrl+c
trap("INT") {
  exit
}
